from typing import Optional


def with_prefix(prefix: Optional[str], value: str) -> str:
    """Add prefix to string.

    If prefix is not empty, this function will add a
    prefix to a string, delimited by a white space.
    """
    prefix = prefix.strip() if prefix else ""
    if prefix:
        return f"{prefix} {value}"
    else:
        return value



STATE_ACTIVE = (True, 2)
TRUE_VALUES = ["+", "Ja", "yes", "Yes", "ja"]
FALSE_VALUES = ["-", "Nein", "no", "No", "nein"]
SEXES = {
    "w": "f",
    "m": "m",
    "weiblich": "f",
    "männlich": "m",
    "female": "f",
    "male": "m",
}



from datetime import date
from typing import Any, Callable, Sequence, Union

import dateparser
from phonenumber_field.phonenumber import PhoneNumber
from phonenumbers import NumberParseException

from aleksis.apps.csv_import.settings import SEXES
from aleksis.core.util.core_helpers import get_site_preferences


class ConverterRegistry:
    """Registry of known conversion functions."""

    def __init__(self):
        self._converters = {}

    def register(self, func: Callable[[Any], Any]) -> Callable[[Any], Any]:
        """Register a conversion function."""
        if func.__name__ in self._converters:
            raise ValueError(f"The converter {func.__name__} is already registered.")
        self._converters[func.__name__] = func

        return func

    def get_from_name(self, name: str) -> Callable[[Any], Any]:
        """Get a conversion function by its name."""
        return self._converters[name]


converter_registry = ConverterRegistry()


@converter_registry.register
def parse_phone_number(value: str) -> Union[PhoneNumber, str]:
    """Parse a phone number."""
    try:
        number = PhoneNumber.from_string(
            value, get_site_preferences()["internationalisation__phone_number_country"]
        )
        if number.is_valid():
            return number
    except NumberParseException:
        pass
    return ""


@converter_registry.register
def parse_sex(value: str) -> str:
    """Parse sex via SEXES dictionary."""
    value = value.lower()
    if value in SEXES:
        return SEXES[value]

    return ""


@converter_registry.register
def parse_date(value: str) -> Union[date, None]:
    """Parse string date."""
    languages_raw = get_site_preferences()["csv_import__date_languages"]
    languages = languages_raw.split(",") if languages_raw else []
    try:
        return dateparser.parse(value, languages=languages).date()
    except (ValueError, AttributeError):
        return None


@converter_registry.register
def parse_comma_separated_data(value: str) -> Sequence[str]:
    """Parse a string with comma-separated data."""
    return list(filter(lambda v: v, value.split(",")))


for str_converter in ("capitalize", "lstrip", "strip", "rstrip", "lower", "upper", "title"):
    converter_registry.register(getattr(str, str_converter))



# Generated by Django 3.2.3 on 2021-05-23 12:52

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0019_fix_uniqueness_per_site'),
        ('csv_import', '0002_importjob'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='importtemplate',
            constraint=models.UniqueConstraint(fields=('site_id', 'name'), name='unique_template_name_per_site'),
        ),
    ]



import os
import re
from typing import Any, Callable, Optional, Sequence, Tuple, Type, Union
from uuid import uuid4

from django.apps import apps
from django.core.files import File
from django.db.models import Model
from django.utils.translation import gettext as _

from aleksis.apps.csv_import.util.class_range_helpers import (
    get_classes_per_grade,
    get_classes_per_short_name,
    parse_class_range,
)
from aleksis.apps.csv_import.util.converters import converter_registry
from aleksis.apps.csv_import.util.import_helpers import with_prefix
from aleksis.core.models import Group, Person, SchoolTerm
from aleksis.core.util.core_helpers import get_site_preferences


class FieldType:
    name: str = ""
    verbose_name: str = ""
    models: Sequence = []
    data_type: type = str
    db_field: str = ""
    converter: Optional[Union[str, Sequence[str]]] = None
    alternative_db_fields: Optional[str] = None
    template: str = ""
    args: Optional[dict] = None

    @classmethod
    def get_data_type(cls) -> type:
        return cls.data_type

    def get_converter(self) -> Callable[[Any], Any]:
        converters_pre = self.get_args().get("converter_pre", [])
        if isinstance(converters_pre, str):
            converters_pre = [converters_pre]
        converters_post = self.get_args().get("converter_post", [])
        if isinstance(converters_post, str):
            converters_post = [converters_post]
        converters = self.get_args().get("converter") or self.converter
        if converters is None:
            converters = []
        elif isinstance(converters, str):
            converters = [converters]
        converters = converters_pre + converters + converters_post

        funcs = [converter_registry.get_from_name(name) for name in converters]

        def _converter_chain(val: Any) -> Any:
            new_val = val
            for func in funcs:
                new_val = func(new_val)
            return new_val

        return _converter_chain

    def get_args(self) -> dict:
        return self.args or {}

    def get_db_field(self) -> str:
        if self.get_args().get("db_field"):
            return self.get_args()["db_field"]
        return self.db_field

    def get_alternative_db_fields(self) -> list[str]:
        if self.get_args().get("alternative_db_fields"):
            return self.get_args()["alterntive_db_fields"]
        return self.alternative_db_fields or []

    def get_column_name(self) -> str:
        """Get column name for use in Pandas structures."""
        if self.get_args().get("column_name"):
            return self.get_args()["column_name"]
        return self.column_name

    def get_template(self) -> str:
        if self.get_args().get("template"):
            return self.get_args()["template"]
        return self.template or ""

    def __init__(self, school_term: SchoolTerm, base_path: str, **kwargs):
        self.school_term = school_term
        self.base_path = os.path.realpath(base_path)
        self.column_name = f"col_{uuid4()}"
        self.args = kwargs


class FieldTypeRegistry:
    def __init__(self):
        self.field_types = {}

    def register(self, field_type: Type[FieldType]):
        """Add new `FieldType` to registry.

        Can be used as decorator, too.
        """
        if field_type.name in self.field_types:
            raise ValueError(f"The field type {field_type.name} is already registered.")
        self.field_types[field_type.name] = field_type

        return field_type

    def get_from_name(self, name: str) -> FieldType:
        """Get `FieldType` by its name."""
        return self.field_types[name]

    @property
    def choices(self) -> Sequence[Tuple[str, str]]:
        """Return choices in Django format."""
        return [(f.name, f.verbose_name) for f in self.field_types.values()]


field_type_registry = FieldTypeRegistry()


@field_type_registry.register
class MatchFieldType(FieldType):
    """Field type for getting an instance."""

    name: str = "match"
    priority: int = 1

    def get_priority(self):
        return self.get_args().get("priority", "") or self.priority


class DirectMappingFieldType(FieldType):
    """Set value directly in DB."""


class ProcessFieldType(FieldType):
    """Field type with custom logic for importing."""

    def process(self, instance: Model, value):
        pass


class RegExFieldType(ProcessFieldType):
    """Field type to apply a regex transformation."""

    data_type = str
    reg_ex: str = ""

    def get_reg_ex(self):
        return self.reg_ex or self.get_args().get("reg_ex", "")

    def process(self, instance: Model, value):
        match = re.fullmatch(self.get_reg_ex(), value)
        if match:
            for key, item in match.groupdict().items():
                setattr(instance, key, item)
            instance.save()


@field_type_registry.register
class UniqueReferenceFieldType(MatchFieldType):
    name = "unique_reference"
    verbose_name = _("Unique reference")
    db_field = "import_ref_csv"
    priority = 10


@field_type_registry.register
class NameFieldType(DirectMappingFieldType):
    name = "name"
    verbose_name = _("Name")
    db_field = "name"
    alternative_db_fields = ["short_name"]


@field_type_registry.register
class FirstNameFieldType(DirectMappingFieldType):
    name = "first_name"
    verbose_name = _("First name")
    db_field = "first_name"


@field_type_registry.register
class LastNameFieldType(DirectMappingFieldType):
    name = "last_name"
    verbose_name = _("Last name")
    db_field = "last_name"


@field_type_registry.register
class AdditionalNameFieldType(DirectMappingFieldType):
    name = "additional_name"
    verbose_name = _("Additional name")
    db_field = "additional_name"


@field_type_registry.register
class ShortNameFieldType(MatchFieldType):
    name = "short_name"
    verbose_name = _("Short name")
    priority = 8
    db_field = "short_name"
    alternative_db_fields = ["name", "first_name", "last_name"]


@field_type_registry.register
class EmailFieldType(MatchFieldType):
    name = "email"
    verbose_name = _("Email")
    db_field = "email"
    priority = 12

    def get_converter(self) -> Optional[Callable]:
        if "email_domain" in self.get_args():

            def add_domain_to_email(value: str) -> str:
                if "@" in value:
                    return value
                else:
                    return f"{value}@{self.get_args()['email_domain']}"

            return add_domain_to_email
        return super().get_converter()


@field_type_registry.register
class DateOfBirthFieldType(DirectMappingFieldType):
    name = "date_of_birth"
    verbose_name = _("Date of birth")
    db_field = "date_of_birth"
    converter = "parse_date"


@field_type_registry.register
class SexFieldType(DirectMappingFieldType):
    name = "sex"
    verbose_name = _("Sex")
    db_field = "sex"
    converter = "parse_sex"


@field_type_registry.register
class StreetFieldType(DirectMappingFieldType):
    name = "street"
    verbose_name = _("Street")
    db_field = "street"


@field_type_registry.register
class HouseNumberFieldType(DirectMappingFieldType):
    name = "housenumber"
    verbose_name = _("Housenumber")
    db_field = "housenumber"


@field_type_registry.register
class StreetAndHouseNumberFieldType(RegExFieldType):
    name = "street_housenumber"
    verbose_name = _("Street and housenumber")
    reg_ex = r"^(?P<street>[\w\s]{3,})\s+(?P<housenumber>\d+\s*[a-zA-Z]*)$"


@field_type_registry.register
class PostalCodeFieldType(DirectMappingFieldType):
    name = "postal_code"
    verbose_name = _("Postal code")
    db_field = "postal_code"


@field_type_registry.register
class PlaceFieldType(DirectMappingFieldType):
    name = "place"
    verbose_name = _("Place")
    db_field = "place"


@field_type_registry.register
class PhoneNumberFieldType(DirectMappingFieldType):
    name = "phone_number"
    verbose_name = _("Phone number")
    db_field = "phone_number"
    converter = "parse_phone_number"


@field_type_registry.register
class MobileNumberFieldType(DirectMappingFieldType):
    name = "mobile_number"
    verbose_name = _("Mobile number")
    db_field = "mobile_number"
    converter = "parse_phone_number"


@field_type_registry.register
class IgnoreFieldType(FieldType):
    name = "ignore"
    verbose_name = _("Ignore data in this field")


@field_type_registry.register
class DepartmentsFieldType(ProcessFieldType):
    name = "departments"
    verbose_name = _("Comma-seperated list of departments")
    converter = "parse_comma_separated_data"

    def process(self, instance: Model, value):
        with_chronos = apps.is_installed("aleksis.apps.chronos")
        if with_chronos:
            Subject = apps.get_model("chronos", "Subject")

        group_type = get_site_preferences()["csv_import__group_type_departments"]
        group_prefix = get_site_preferences()["csv_import__group_prefix_departments"]

        groups = []
        for subject_name in value:
            if with_chronos:
                # Get department subject
                subject, __ = Subject.objects.get_or_create(
                    short_name=subject_name, defaults={"name": subject_name}
                )
                name = subject.name
                short_name = subject.short_name
            else:
                name = subject_name
                short_name = subject_name

            # Get department group
            group, __ = Group.objects.get_or_create(
                group_type=group_type,
                short_name=short_name,
                defaults={"name": with_prefix(group_prefix, name)},
            )
            if with_chronos:
                group.subject = subject
            group.save()

            groups.append(group)

        instance.member_of.add(*groups)


@field_type_registry.register
class GroupSubjectByShortNameFieldType(ProcessFieldType):
    name = "group_subject_short_name"
    verbose_name = _("Short name of the subject")

    def process(self, instance: Model, value):
        with_chronos = apps.is_installed("aleksis.apps.chronos")
        if with_chronos:
            Subject = apps.get_model("chronos", "Subject")
            subject, __ = Subject.objects.get_or_create(short_name=value, defaults={"name": value})
            instance.subject = subject
            instance.save()


@field_type_registry.register
class ClassRangeFieldType(ProcessFieldType):
    name = "class_range"
    verbose_name = _("Class range (e. g. 7a-d)")

    def __init__(self, school_term: SchoolTerm, base_path: str, **kwargs):
        # Prefetch class groups
        self.classes_per_short_name = get_classes_per_short_name(school_term)
        self.classes_per_grade = get_classes_per_grade(self.classes_per_short_name.keys())

        super().__init__(school_term, base_path, **kwargs)

    def process(self, instance: Model, value):
        classes = parse_class_range(
            self.classes_per_short_name,
            self.classes_per_grade,
            value,
        )
        instance.parent_groups.set(classes)


@field_type_registry.register
class PrimaryGroupByShortNameFieldType(ProcessFieldType):
    name = "primary_group_short_name"
    verbose_name = _("Short name of the person's primary group")

    def process(self, instance: Model, value):
        group, __ = Group.objects.get_or_create(
            short_name=value, school_term=self.school_term, defaults={"name": value}
        )
        instance.primary_group = group
        instance.member_of.add(group)
        instance.save()


@field_type_registry.register
class PrimaryGroupOwnerByShortNameFieldType(ProcessFieldType):
    name = "primary_group_owner_short_name"
    verbose_name = _("Short name of an owner of the person's primary group")

    def process(self, instance: Model, value):
        if instance.primary_group:
            owners = Person.objects.filter(short_name=value)
            instance.primary_group.owners.set(owners)


@field_type_registry.register
class GroupOwnerByShortNameFieldType(ProcessFieldType):
    name = "group_owner_short_name"
    verbose_name = _("Short name of a single group owner")

    def process(self, instance: Model, short_name: str):
        group_owner, __ = Person.objects.get_or_create(
            short_name=short_name,
            defaults={"first_name": "?", "last_name": short_name},
        )
        if self.get_args().get("clear", False):
            instance.owners.set([group_owner])
        else:
            instance.owners.add(group_owner)


@field_type_registry.register
class GroupMembershipByShortNameFieldType(ProcessFieldType):
    name = "group_membership_short_name"
    verbose_name = _("Short name of the group the person is a member of")

    def process(self, instance: Model, short_name: str):
        try:
            group = Group.objects.get(short_name=short_name, school_term=self.school_term)
            instance.member_of.add(group)
        except Group.DoesNotExist:
            pass


@field_type_registry.register
class ParentGroupByShortNameFieldType(ProcessFieldType):
    name = "parent_group_short_name"
    verbose_name = _("Short name of the group's parent group")

    def process(self, instance: Model, value):
        group, __ = Group.objects.get_or_create(
            short_name=value, school_term=self.school_term, defaults={"name": value}
        )
        instance.parent_groups.add(group)
        instance.save()


@field_type_registry.register
class MemberOfByNameFieldType(ProcessFieldType):
    name = "member_of_by_name"
    verbose_name = _("Name of a group the person is a member of")

    def process(self, instance: Model, value):
        group, __ = Group.objects.get_or_create(name=value, school_term=self.school_term)
        instance.member_of.add(group)
        instance.save()


@field_type_registry.register
class ChildByUniqueReference(ProcessFieldType):
    name = "child_by_unique_reference"
    verbose_name = _("Child by unique reference (from students import)")

    def process(self, instance: Model, value):
        child = Person.objects.get(import_ref_csv=value)
        instance.children.add(child)


@field_type_registry.register
class FileFieldType(ProcessFieldType):
    """Field type that stores a referenced file on a file field."""

    def process(self, instance: Model, value: str):
        # Test path for unwanted path traversal
        abs_path = os.path.realpath(value)
        if not self.base_path == os.path.commonpath((self.base_path, abs_path)):
            raise ValueError(f"Disallowed path traversal importing file from {value}")

        # Get target FileField and save content
        field_name = self.get_db_field()
        file_field = getattr(instance, field_name)
        file_field.save(os.path.basename(abs_path), File(open(abs_path, "rb")))

        instance.save()


@field_type_registry.register
class AvatarFieldType(FileFieldType):
    name = "avatar"
    db_field = "avatar"


@field_type_registry.register
class PhotoFieldType(FileFieldType):
    name = "photo"
    db_field = "photo"



from django.contrib.contenttypes.models import ContentType

import pytest

from aleksis.apps.csv_import.field_types import ShortNameFieldType, UniqueReferenceFieldType
from aleksis.apps.csv_import.models import ImportTemplate
from aleksis.core.models import Person

pytestmark = pytest.mark.django_db


def test_import_template_str():
    template = ImportTemplate.objects.create(
        content_type=ContentType.objects.get_for_model(Person),
        name="foo",
        verbose_name="Bar",
    )
    assert str(template) == "Bar"


def test_import_template_field():
    template = ImportTemplate.objects.create(
        content_type=ContentType.objects.get_for_model(Person),
        name="foo",
        verbose_name="Bar",
    )
    field_0 = template.fields.create(field_type=UniqueReferenceFieldType.name, index=0)
    field_1 = template.fields.create(field_type=ShortNameFieldType.name, index=1)

    assert field_0.field_type_class == UniqueReferenceFieldType
    assert field_1.field_type_class == ShortNameFieldType

    assert template.fields.count() == 2



from django.utils.translation import gettext_lazy as _

MENUS = {
    "DATA_MANAGEMENT_MENU": [
        {
            "name": _("CSV import"),
            "url": "csv_import",
            "validators": [
                (
                    "aleksis.core.util.predicates.permission_validator",
                    "csv_import.import_data_rule",
                ),
            ],
        },
        {
            "name": _("CSV import templates"),
            "url": "import_templates",
            "validators": [
                (
                    "aleksis.core.util.predicates.permission_validator",
                    "csv_import.view_importtemplate_rule",
                ),
            ],
        },
    ]
}



from django import forms
from django.core.exceptions import ValidationError
from django.core.validators import FileExtensionValidator
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from ruamel.yaml import YAML, YAMLError

from aleksis.apps.csv_import.models import ImportTemplate
from aleksis.core.models import SchoolTerm


class CSVUploadForm(forms.Form):
    csv = forms.FileField(label=_("CSV file"))
    school_term = forms.ModelChoiceField(
        queryset=SchoolTerm.objects.all(),
        label=_("Related school term"),
    )
    template = forms.ModelChoiceField(
        queryset=ImportTemplate.objects.all(), label=_("Import template")
    )

    def __init__(self, *args, **kwargs):
        try:
            school_terms = SchoolTerm.objects.on_day(timezone.now().date())
            kwargs["initial"] = {"school_term": school_terms[0] if school_terms.exists() else None}
        except SchoolTerm.DoesNotExist:
            pass
        super().__init__(*args, **kwargs)


class ImportTemplateUploadForm(forms.Form):
    template = forms.FileField(
        label=_("CSV template"), validators=[FileExtensionValidator(["yml", "yaml"])]
    )

    def clean_template(self):
        """Check if the template is a valid YAML file."""
        file = self.cleaned_data["template"]
        try:
            yaml = YAML(typ="safe")
            template_defs = yaml.load(file)

            default_template_names = ImportTemplate.objects.filter(from_default=True).values_list(
                "name", flat=True
            )
            if set(template_defs.keys()) & set(default_template_names):
                raise ValidationError(
                    _("You cannot import a template with the same name as a default template.")
                )
            return template_defs
        except YAMLError as e:
            raise ValidationError(_("Invalid YAML file uploaded: \n {}").format(e))



from aleksis.apps.csv_import.models import ImportJob
from aleksis.core.util.celery_progress import ProgressRecorder, recorded_task

from .util.process import import_csv as _import_csv


@recorded_task
def import_csv(
    import_job: int,
    recorder: ProgressRecorder,
) -> None:
    import_job = ImportJob.objects.get(pk=import_job)
    _import_csv(import_job, recorder)



# Generated by Django 3.2.10 on 2021-12-30 10:29

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0019_fix_uniqueness_per_site'),
        ('csv_import', '0003_fix_uniqueness_per_site'),
    ]

    operations = [
        migrations.AddField(
            model_name='importtemplatefield',
            name='args',
            field=models.JSONField(default={}, verbose_name='Optional arguments passed to field type'),
        ),
    ]



# Generated by Django 3.2.10 on 2022-01-01 22:30

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('csv_import', '0004_args'),
    ]

    operations = [
        migrations.AddField(
            model_name='importtemplate',
            name='has_index_col',
            field=models.BooleanField(default=False, verbose_name='Has the CSV file an own index column?'),
        ),
    ]



from django.contrib.contenttypes.models import ContentType

import pytest

from aleksis.apps.csv_import.field_types import FirstNameFieldType, LastNameFieldType
from aleksis.apps.csv_import.models import ImportTemplate
from aleksis.core.models import Person

pytestmark = pytest.mark.django_db


def test_converters_none():
    template = ImportTemplate.objects.create(
        content_type=ContentType.objects.get_for_model(Person),
        name="foo",
        verbose_name="Bar",
    )

    field_0 = template.fields.create(field_type=FirstNameFieldType.name, index=0)
    field_1 = template.fields.create(field_type=LastNameFieldType.name, index=1)

    field_0_cls = field_0.field_type_class(None, "", **field_0.args)
    field_1_cls = field_1.field_type_class(None, "", **field_1.args)

    field_0_chain = field_0_cls.get_converter()
    field_1_chain = field_1_cls.get_converter()

    convert_in = " foo "
    convert_out_0 = field_0_chain(convert_in)
    convert_out_1 = field_1_chain(convert_in)

    assert convert_out_0 == convert_in
    assert convert_out_1 == convert_in


def test_converters_one():
    template = ImportTemplate.objects.create(
        content_type=ContentType.objects.get_for_model(Person),
        name="foo",
        verbose_name="Bar",
    )

    field_0 = template.fields.create(
        field_type=FirstNameFieldType.name, index=0, args={"converter_pre": "lstrip"}
    )
    field_1 = template.fields.create(
        field_type=LastNameFieldType.name, index=1, args={"converter_post": "rstrip"}
    )

    field_0_cls = field_0.field_type_class(None, "", **field_0.args)
    field_1_cls = field_1.field_type_class(None, "", **field_1.args)

    field_0_chain = field_0_cls.get_converter()
    field_1_chain = field_1_cls.get_converter()

    convert_in = " foo "
    convert_out_0 = field_0_chain(convert_in)
    convert_out_1 = field_1_chain(convert_in)

    assert convert_out_0 == convert_in.lstrip()
    assert convert_out_1 == convert_in.rstrip()


def test_converters_multiple():
    template = ImportTemplate.objects.create(
        content_type=ContentType.objects.get_for_model(Person),
        name="foo",
        verbose_name="Bar",
    )

    field_0 = template.fields.create(
        field_type=FirstNameFieldType.name, index=0, args={"converter_pre": ["lstrip", "rstrip"]}
    )
    field_1 = template.fields.create(
        field_type=LastNameFieldType.name,
        index=1,
        args={"converter_pre": "rstrip", "converter_post": "lstrip"},
    )

    field_0_cls = field_0.field_type_class(None, "", **field_0.args)
    field_1_cls = field_1.field_type_class(None, "", **field_1.args)

    field_0_chain = field_0_cls.get_converter()
    field_1_chain = field_1_cls.get_converter()

    convert_in = " foo "
    convert_out_0 = field_0_chain(convert_in)
    convert_out_1 = field_1_chain(convert_in)

    assert convert_out_0 == convert_in.strip()
    assert convert_out_1 == convert_in.strip()



import os

from ruamel.yaml import YAML

from .models import ImportTemplate


def update_or_create_templates_from_yaml(yaml_file: str):
    """Update or create import templates from a yaml file."""
    with open(yaml_file, "r") as f:
        yaml = YAML(typ="safe")
        template_defs = yaml.load(f)

    ImportTemplate.update_or_create_templates(template_defs, from_default=True)


def update_or_create_default_templates():
    """Update or create default import templates."""
    update_or_create_templates_from_yaml(
        os.path.join(os.path.dirname(__file__), "default_templates.yaml")
    )



# Generated by Django 3.2.11 on 2022-01-07 16:26

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('csv_import', '0005_optional_index_col'),
    ]

    operations = [
        migrations.AddField(
            model_name='importtemplate',
            name='from_default',
            field=models.BooleanField(default=False),
        ),
    ]



from django.utils.translation import gettext as _

from jsonstore import CharField

from aleksis.core.mixins import ExtensibleModel

for model in ExtensibleModel.__subclasses__():
    model.field(import_ref_csv=CharField(verbose_name=_("CSV import reference"), blank=True))



# Generated by Django 3.1.3 on 2020-11-08 16:14

import django.contrib.sites.managers
import django.db.models.deletion
from django.db import migrations, models

import aleksis.apps.csv_import.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("sites", "0002_alter_domain_unique"),
        ("core", "0003_drop_image_cropping"),
    ]

    operations = [
        migrations.CreateModel(
            name="ImportTemplate",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("extended_data", models.JSONField(default=dict, editable=False)),
                (
                    "name",
                    models.CharField(max_length=255, unique=True, verbose_name="Name"),
                ),
                ("verbose_name", models.CharField(max_length=255, verbose_name="Name")),
                (
                    "has_header_row",
                    models.BooleanField(
                        default=True, verbose_name="Has the CSV file an own header row?"
                    ),
                ),
                (
                    "separator",
                    models.CharField(
                        default=",",
                        help_text="For whitespace use \\\\s+, for tab \\\\t",
                        max_length=255,
                        verbose_name="CSV separator",
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                        verbose_name="Content type",
                    ),
                ),
                (
                    "group",
                    models.ForeignKey(
                        blank=True,
                        help_text="If imported objects are persons, they all will be members of this group after import.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="core.group",
                        verbose_name="Base group",
                    ),
                ),
                (
                    "group_type",
                    models.ForeignKey(
                        blank=True,
                        help_text="If imported objects are groups, they all will get this group type after import.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="core.grouptype",
                        verbose_name="Gruppentyp",
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        default=1,
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="sites.site",
                    ),
                ),
            ],
            options={
                "verbose_name": "Import template",
                "verbose_name_plural": "Import templates",
                "ordering": ["name"],
            },
            managers=[("objects", django.contrib.sites.managers.CurrentSiteManager()),],
        ),
        migrations.CreateModel(
            name="ImportTemplateField",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("extended_data", models.JSONField(default=dict, editable=False)),
                ("index", models.IntegerField(verbose_name="Index")),
                (
                    "field_type",
                    models.CharField(
                        choices=[
                            ("unique_reference", "Unique reference"),
                            ("name", "Name"),
                            ("first_name", "Vorname"),
                            ("last_name", "Nachname"),
                            ("additional_name", "Additional name"),
                            ("short_name", "Kurzname"),
                            ("email", "E-Mail-Adresse"),
                            ("date_of_birth", "Geburtsdatum"),
                            ("sex", "Geschlecht"),
                            ("street", "Straße"),
                            ("housenumber", "Housenumber"),
                            ("postal_code", "Postleitzahl"),
                            ("place", "Ort"),
                            ("phone_number", "Telefonnummer"),
                            ("mobile_number", "Mobile number"),
                            ("ignore", "Ignore data in this field"),
                            ("departments", "Comma-seperated list of departments"),
                            ("group_subject_short_name", "Short name of the subject"),
                            ("class_range", "Class range (e. g. 7a-d)"),
                            (
                                "primary_group_short_name",
                                "Short name of the person's primary group",
                            ),
                            (
                                "group_owner_short_name",
                                "Short name of a single group owner",
                            ),
                            (
                                "group_membership_short_name",
                                "Short name of the group the person is a member of",
                            ),
                            (
                                "child_by_unique_reference",
                                "Child by unique reference (from students import)",
                            ),
                        ],
                        max_length=255,
                        verbose_name="Field type",
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        default=1,
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="sites.site",
                    ),
                ),
                (
                    "template",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="fields",
                        to="csv_import.importtemplate",
                        verbose_name="Import template",
                    ),
                ),
            ],
            options={
                "verbose_name": "Import template field",
                "verbose_name_plural": "Import template fields",
                "ordering": ["template", "index"],
                "unique_together": {("template", "index")},
            },
            managers=[("objects", django.contrib.sites.managers.CurrentSiteManager()),],
        ),
    ]



from django.core.exceptions import ImproperlyConfigured
from django.db import OperationalError, ProgrammingError

from aleksis.core.util.apps import AppConfig


class CSVImportConfig(AppConfig):
    name = "aleksis.apps.csv_import"
    dist_name = "AlekSIS-App-CSVImport"
    verbose_name = "AlekSIS — CSV import"

    urls = {
        "Repository": "https://edugit.org/AlekSIS/official/AlekSIS-App-CSVImport/",
    }
    licence = "EUPL-1.2+"
    copyright_info = (
        ([2019, 2020, 2022], "Dominik George", "dominik.george@teckids.org"),
        ([2020, 2021], "Jonathan Weth", "dev@jonathanweth.de"),
        ([2019], "mirabilos", "thorsten.glaser@teckids.org"),
        ([2019], "Tom Teichler", "tom.teichler@teckids.org"),
    )

    def ready(self):
        super().ready()

        # Create default import templates
        try:
            from aleksis.apps.csv_import.default_templates import (  # noqa
                update_or_create_default_templates,
            )

            update_or_create_default_templates()
        except (ProgrammingError, OperationalError, ImproperlyConfigured) as e:
            # Catch if there are no migrations yet
            pass  # noqa



import os.path

from django.core.management.base import BaseCommand
from django.utils.translation import gettext as _

from aleksis.core.models import SchoolTerm
from aleksis.core.util import messages

from ...models import ImportJob, ImportTemplate
from ...tasks import import_csv
from ...util.process import import_csv as import_csv_base


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "csv_path",
            help=_("Path to CSV file"),
        )
        parser.add_argument(
            "template",
            help=_("Name of import template which should be used"),
        )
        parser.add_argument(
            "--school-term",
            help=_("ID of School term to which the imported objects should belong"),
            type=int,
        )
        parser.add_argument(
            "--background",
            action="store_true",
            help="Run import job in background using Celery",
        )

    def handle(self, *args, **options):
        template_name = options["template"]
        try:
            template = ImportTemplate.objects.get(name=template_name)
        except ImportTemplate.DoesNotExist:
            messages.error(None, _("The provided template does not exist."))
            return

        try:
            pk = options["school_term"]
            school_term = SchoolTerm.objects.get(pk=pk)
        except SchoolTerm.DoesNotExist:
            school_term = None

        filename = os.path.split(options["csv_path"])[1]

        import_job = ImportJob(template=template, school_term=school_term)
        with open(options["csv_path"], "r") as csv_file:
            import_job.data_file.save(filename, csv_file)
        import_job.save()

        if options["background"]:
            import_csv.delay(import_job.pk)
        else:
            import_csv_base(import_job)



import logging
import os
from glob import glob
from tempfile import TemporaryDirectory
from typing import Optional
from zipfile import ZipFile

from django.conf import settings
from django.contrib import messages
from django.core.exceptions import ValidationError
from django.db import transaction
from django.template import Context, Template
from django.utils.translation import gettext as _

import chardet
import pandas
from pandas.errors import ParserError
from tqdm import tqdm

from aleksis.apps.csv_import.field_types import (
    DirectMappingFieldType,
    MatchFieldType,
    ProcessFieldType,
)
from aleksis.apps.csv_import.settings import FALSE_VALUES, TRUE_VALUES
from aleksis.core.models import Group, Person
from aleksis.core.util.celery_progress import ProgressRecorder
from aleksis.core.util.core_helpers import process_custom_context_processors

from ..models import ImportJob


@transaction.atomic
def import_csv(
    import_job: ImportJob,
    recorder: Optional[ProgressRecorder] = None,
) -> None:
    """Import one CSV/ZIP file from a job."""
    # We work in a temporary directory locally to allow more import scenarios,
    # like ZIP files that need to be preprocessed
    with TemporaryDirectory() as temp_dir:
        # Get the job and the target of the import
        template = import_job.template
        model = template.content_type.model_class()
        school_term = import_job.school_term

        # Dissect template definition
        # These structures will be filled with information for columns
        data_types = {}
        converters = {}
        field_types = {}
        virtual_fields = []
        for field in template.fields.all():
            # Get field type and prepare for import
            field_type = field.field_type_class(school_term, temp_dir, **field.args)
            column_name = field_type.get_column_name()

            if not field.virtual:
                # Get data type and conversion rules, if any,
                # to be passed to Pandas
                data_types[column_name] = field_type.get_data_type()
                if field_type.get_converter():
                    converters[column_name] = field_type.get_converter()
            else:
                # Mark field as virtual so as to not handle with Pandas
                virtual_fields.append(column_name)
                field_type.template = field.virtual_tmpl

            field_types[column_name] = field_type

        # Determine whether the data file is a plain CSV or an archive
        if import_job.data_file.name.endswith(".zip"):
            # Unpack to temporary directory
            with ZipFile(import_job.data_file.open("rb")) as zip_file:
                zip_file.extractall(temp_dir)

            # Find all CSV files
            csv_names = glob(os.path.join(temp_dir, "*.csv")) + glob(
                os.path.join(temp_dir, "**", "*.csv")
            )
        else:
            # Copy CSV file to temporary directory verbatim
            temp_csv_name = os.path.join(temp_dir, os.path.basename(import_job.data_file.name))
            with open(temp_csv_name, "wb") as temp_csv:
                temp_csv.write(import_job.data_file.open("rb").read())

            csv_names = [temp_csv_name]

        for csv_name in csv_names:
            # chdir() to current CSV file directory; needed for finding
            # related files if importing file columns from zipped CSVs
            os.chdir(os.path.dirname(csv_name))

            # Guess encoding first
            with open(csv_name, "rb") as csv:
                encoding = chardet.detect(csv.read())["encoding"]

            with open(csv_name, newline="", encoding=encoding) as csv:
                try:
                    # Use discovered column configurations for one-off Pandas loading
                    data = pandas.read_csv(
                        csv,
                        sep=template.parsed_separator,
                        names=data_types.keys(),
                        header=0 if template.has_header_row else None,
                        index_col=template.has_index_col,
                        dtype=data_types,
                        usecols=lambda k: not k.startswith("_"),
                        keep_default_na=False,
                        converters=converters,
                        quotechar=template.parsed_quotechar,
                        encoding=encoding,
                        true_values=TRUE_VALUES,
                        false_values=FALSE_VALUES,
                    )
                except ParserError as e:
                    msg = _("There was an error while parsing the CSV file:\n{}").format(e)
                    if recorder:
                        recorder.add_message(messages.ERROR, msg)
                    else:
                        logging.error(msg)
                    continue

                # Exclude all empty rows
                data = data.where(data.notnull(), None)

                all_ok = True
                inactive_refs = []
                created_count = 0

                data_as_dict = data.transpose().to_dict().values()

                iterator = recorder.iterate(data_as_dict) if recorder else tqdm(data_as_dict)
                for row in iterator:
                    # Generate virtual and post-processed field data
                    for column_name, field_type in field_types.items():
                        # Generate field using a Django template string, and the row as context
                        tmpl_str = field_type.get_template()
                        if not tmpl_str:
                            continue
                        tmpl = Template(tmpl_str)
                        ctx = Context(row)
                        ctx.update(
                            process_custom_context_processors(
                                settings.NON_REQUEST_CONTEXT_PROCESSORS
                            )
                        )
                        data = tmpl.render(ctx).strip()

                        if column_name in virtual_fields:
                            # Post-process virtual fields using converter
                            data = field_type.get_converter()(data)

                        # Store
                        row[column_name] = data

                    # Build dict with all fields that should be directly updated
                    update_dict = {}
                    for key, value in row.items():
                        if key in field_types:
                            field_type = field_types[key]
                            if isinstance(field_type, DirectMappingFieldType):
                                update_dict[field_type.get_db_field()] = value

                    # Set group type for imported groups if defined in template globally
                    if template.group_type and model == Group:
                        update_dict["group_type"] = template.group_type

                    # Determine available fields for finding existing instances
                    get_dict = {}
                    match_field_priority = 0
                    for (column_name, field_type,) in sorted(
                        filter(lambda f: isinstance(f[1], MatchFieldType), field_types.items()),
                        key=lambda f: f[1].priority,
                    ):
                        if match_field_priority and match_field_priority < field_type.priority:
                            # We found a match field, but with less important priority than
                            # those before, so we write data from the field
                            update_dict[field_type.get_db_field()] = row[column_name]
                        elif column_name in row:
                            get_dict[field_type.get_db_field()] = row[column_name]
                            match_field_priority = field_type.priority

                    if not match_field_priority:
                        raise ValueError(_("Missing unique reference or other matching fields."))

                    # Set alternatives for some fields
                    for column_name, field_type in field_types.items():
                        for alternative_db_field in field_type.get_alternative_db_fields():
                            origin_db_field = field_type.get_db_field()
                            if (
                                hasattr(model, alternative_db_field)
                                and alternative_db_field not in update_dict
                            ):
                                if origin_db_field in update_dict:
                                    update_dict[alternative_db_field] = update_dict[origin_db_field]
                                elif origin_db_field in get_dict:
                                    update_dict[alternative_db_field] = get_dict[origin_db_field]

                    # Set school term globally if model is school term related
                    if hasattr(model, "school_term") and school_term:
                        get_dict["school_term"] = school_term

                    created = False
                    try:
                        get_dict["defaults"] = update_dict
                        instance, created = model.objects.update_or_create(**get_dict)
                        if created:
                            created_count += 1

                        # Process field types with custom logic
                        for column_name, field_type in filter(
                            lambda f: isinstance(f[1], ProcessFieldType), field_types.items()
                        ):
                            try:
                                field_type.process(instance, row[column_name])
                            except RuntimeError as e:
                                if recorder:
                                    recorder.add_message(messages.ERROR, str(e))
                                else:
                                    logging.error(str(e))

                        # Add current instance to group if import defines a target group for persons
                        if template.group and isinstance(instance, Person):
                            instance.member_of.add(template.group)
                    except (
                        ValueError,
                        ValidationError,
                        model.MultipleObjectsReturned,
                        model.DoesNotExist,
                    ) as e:
                        msg = _("Failed to import {verbose_name} {row}:\n{e}").format(
                            verbose_name=model._meta.verbose_name, row=row, e=e
                        )
                        if recorder:
                            recorder.add_message(
                                messages.ERROR,
                                msg,
                            )
                        else:
                            logging.error(msg)
                            all_ok = False

                if created_count:
                    msg = _("{created_count} {verbose_name} were newly created.").format(
                        created_count=created_count, verbose_name=model._meta.verbose_name_plural
                    )
                    if recorder:
                        recorder.add_message(messages.SUCCESS, msg)
                    else:
                        logging.info(msg)

                if all_ok:
                    msg = _("All {verbose_name} were imported successfully.").format(
                        verbose_name=model._meta.verbose_name_plural
                    )
                    if recorder:
                        recorder.add_message(messages.SUCCESS, msg)
                    else:
                        logging.info(msg)
                else:
                    msg = _("Some {verbose_name} failed to be imported.").format(
                        verbose_name=model._meta.verbose_name_plural
                    )
                    if recorder:
                        recorder.add_message(messages.WARNING, msg)
                    else:
                        logging.warning(msg)



from aleksis.apps.csv_import.util.import_helpers import with_prefix
from aleksis.core.models import Group, GroupType, Person


def test_with_prefix():
    assert with_prefix("Foo", "Bar") == "Foo Bar"
    assert with_prefix("", "Bar") == "Bar"
    assert with_prefix(None, "Bar") == "Bar"
    assert with_prefix(None, "") == ""
    assert with_prefix("", "") == ""



from django_tables2 import BooleanColumn, Column, Table


class ImportTemplateTable(Table):
    """Table to list import templates."""

    content_type = Column()
    name = Column()
    verbose_name = Column()
    has_header_row = BooleanColumn()
    has_index_col = BooleanColumn()

    class Meta:
        attrs = {"class": "highlight"}



import codecs
from typing import Any, Dict, Sequence, Tuple

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Model
from django.utils.translation import gettext as _

from aleksis.core.mixins import ExtensibleModel
from aleksis.core.models import Group, GroupType, SchoolTerm

from .field_types import FieldType, field_type_registry


class ImportTemplate(ExtensibleModel):
    from_default = models.BooleanField(default=False)
    content_type = models.ForeignKey(
        ContentType,
        models.CASCADE,
        verbose_name=_("Content type"),
    )
    name = models.CharField(max_length=255, verbose_name=_("Name"))
    verbose_name = models.CharField(max_length=255, verbose_name=_("Name"))

    has_header_row = models.BooleanField(
        default=True, verbose_name=_("Has the CSV file an own header row?")
    )
    has_index_col = models.BooleanField(
        default=False, verbose_name=_("Has the CSV file an own index column?")
    )
    separator = models.CharField(
        max_length=255,
        default=",",
        verbose_name=_("CSV separator"),
        help_text=_("For whitespace use \\\\s+, for tab \\\\t"),
    )
    quotechar = models.CharField(
        max_length=255,
        default='"',
        verbose_name=_("CSV quote character"),
    )

    group = models.ForeignKey(
        Group,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name=_("Base group"),
        help_text=_(
            "If imported objects are persons, they all will be members of this group after import."
        ),
    )
    group_type = models.ForeignKey(
        GroupType,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        verbose_name=_("Group type"),
        help_text=_(
            "If imported objects are groups, they all will get this group type after import."
        ),
    )

    @property
    def parsed_separator(self):
        return codecs.escape_decode(bytes(self.separator, "utf-8"))[0].decode("utf-8")

    @property
    def parsed_quotechar(self):
        return codecs.escape_decode(bytes(self.quotechar, "utf-8"))[0].decode("utf-8")

    def save(self, *args, **kwargs):
        if not self.content_type.model == "person":
            self.group = None
        if not self.content_type.model == "group":
            self.group_type = None
        super().save(*args, **kwargs)

    @classmethod
    def update_or_create(
        cls,
        model: Model,
        name: str,
        verbose_name: str,
        extra_args: dict,
        fields: Sequence[Tuple[FieldType, Dict[str, Any], bool, str]],
    ):
        """Update or create an import template in database."""
        ct = ContentType.objects.get_for_model(model)
        template, updated = cls.objects.update_or_create(
            name=name,
            defaults={"verbose_name": verbose_name, "content_type": ct, **extra_args},
        )

        i = 0
        for i, field in enumerate(fields):
            field_type, args, virtual, virtual_tmpl = field
            ImportTemplateField.objects.update_or_create(
                template=template,
                index=i,
                defaults={
                    "field_type": field_type.name,
                    "args": args,
                    "virtual": virtual,
                    "virtual_tmpl": virtual_tmpl,
                },
            )

        ImportTemplateField.objects.filter(template=template, index__gt=i).delete()

        return template

    @classmethod
    def update_or_create_templates(cls, template_defs, from_default: bool = False):
        """Update or create import templates."""
        for name, defs in template_defs.items():
            model = apps.get_model(defs["model"])

            fields = []
            for field_definition in defs["fields"]:
                if isinstance(field_definition, str):
                    field_type = field_type_registry.get_from_name(field_definition)
                    virtual = False
                    template = ""
                    args = {}
                else:
                    field_type = field_type_registry.get_from_name(
                        field_definition.pop("field_type")
                    )
                    virtual = field_definition.pop("virtual", False)
                    template = field_definition.pop("template", "")
                    args = field_definition

                fields.append((field_type, args, virtual, template))

            extra_args = defs.get("extra_args", {})
            extra_args["from_default"] = from_default

            cls.update_or_create(
                model,
                name=name,
                verbose_name=defs.get("verbose_name", ""),
                extra_args=extra_args,
                fields=fields,
            )

    def __str__(self):
        return self.verbose_name

    class Meta:
        ordering = ["name"]
        verbose_name = _("Import template")
        verbose_name_plural = _("Import templates")
        constraints = [
            models.UniqueConstraint(
                fields=("site_id", "name"), name="unique_template_name_per_site"
            ),
        ]


class ImportTemplateField(ExtensibleModel):
    index = models.IntegerField(verbose_name=_("Index"))
    template = models.ForeignKey(
        ImportTemplate,
        models.CASCADE,
        verbose_name=_("Import template"),
        related_name="fields",
    )
    field_type = models.CharField(
        max_length=255,
        verbose_name=_("Field type"),
        choices=field_type_registry.choices,
    )
    args = models.JSONField(verbose_name=_("Optional arguments passed to field type"), default={})

    virtual = models.BooleanField(verbose_name=_("Virtual field"), default=False, null=False)
    virtual_tmpl = models.TextField(
        verbose_name=_("Django template to generate virtual field from"), blank=True
    )

    @property
    def field_type_class(self):
        field_type = field_type_registry.get_from_name(self.field_type)
        return field_type

    def clean(self):
        """Validate correct usage of field types."""
        model = self.template.content_type.model_class()
        if (
            self.field_type not in field_type_registry.allowed_field_types_for_models[model]
            and self.field_type not in field_type_registry.allowed_field_types_for_all_models
        ):
            raise ValidationError(_("You are not allowed to use this field type in this model."))

    class Meta:
        ordering = ["template", "index"]
        unique_together = ["template", "index"]
        verbose_name = _("Import template field")
        verbose_name_plural = _("Import template fields")


class ImportJob(ExtensibleModel):
    """Job definition for one import, to track import history and files."""

    template = models.ForeignKey(
        ImportTemplate,
        on_delete=models.CASCADE,
        verbose_name=_("Import template"),
        related_name="import_jobs",
    )
    data_file = models.FileField(upload_to="csv_import/")
    school_term = models.ForeignKey(
        SchoolTerm,
        on_delete=models.CASCADE,
        verbose_name=_("School term"),
        related_name="import_jobs",
        blank=True,
        null=True,
    )

    class Meta:
        verbose_name = _("Import job")
        verbose_name_plural = _("Import jobs")



from django.utils.translation import gettext_lazy as _

from dynamic_preferences.preferences import Section
from dynamic_preferences.types import ModelChoicePreference, StringPreference

from aleksis.core.models import Group, GroupType
from aleksis.core.registries import site_preferences_registry

csv_import = Section("csv_import", verbose_name=_("CSV import"))


@site_preferences_registry.register
class GroupTypeDepartments(ModelChoicePreference):
    section = csv_import
    name = "group_type_departments"
    required = False
    default = None
    model = GroupType
    verbose_name = _("Group type for department groups")
    help_text = _("If you leave it empty, no group type will be used.")


@site_preferences_registry.register
class GroupPrefixDepartments(StringPreference):
    section = csv_import
    name = "group_prefix_departments"
    default = ""
    verbose_name = _("Prefix for long names of department groups")
    help_text = _("If you leave it empty, no prefix will be added.")


@site_preferences_registry.register
class GroupGuardians(ModelChoicePreference):
    section = csv_import
    name = "group_guardians"
    required = False
    default = None
    model = Group
    verbose_name = _("Group for guardians")
    help_text = _("If you leave it empty, no group will be used.")


@site_preferences_registry.register
class DateLanguages(StringPreference):
    section = csv_import
    name = "date_languages"
    required = False
    default = ""
    verbose_name = _("Languages for date parsing")
    help_text = _("e. g. en,es,zh-Hant")



from rules import add_perm

from aleksis.core.util.predicates import has_global_perm, has_person

import_data_predicate = has_person & has_global_perm("csv_import.import_data")
add_perm("csv_import.import_data_rule", import_data_predicate)

view_importtemplate_predicate = has_person & has_global_perm("csv_import.view_importtemplate")
add_perm("csv_import.view_importtemplate_rule", view_importtemplate_predicate)

upload_importtemplate_predicate = has_person & (
    has_global_perm("csv_import.add_importtemplate")
    | has_global_perm("csv_import.change_importtemplate")
)
add_perm("csv_import.upload_importtemplate_rule", upload_importtemplate_predicate)



# Generated by Django 3.2.11 on 2022-01-29 22:44

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('csv_import', '0007_virtual_template_field'),
    ]

    operations = [
        migrations.AddField(
            model_name='importtemplate',
            name='quotechar',
            field=models.CharField(default='"', max_length=255, verbose_name='CSV quote character'),
        ),
    ]



from django.contrib import messages
from django.http import HttpRequest, HttpResponse
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.generic import FormView, ListView

from django_tables2 import SingleTableMixin
from rules.contrib.views import PermissionRequiredMixin, permission_required

from aleksis.core.util.celery_progress import render_progress_page

from .forms import CSVUploadForm, ImportTemplateUploadForm
from .models import ImportJob, ImportTemplate
from .tables import ImportTemplateTable
from .tasks import import_csv


@permission_required("csv_import.import_data_rule")
def csv_import(request: HttpRequest) -> HttpResponse:
    context = {}

    upload_form = CSVUploadForm()

    if request.method == "POST":
        upload_form = CSVUploadForm(request.POST, request.FILES)

        if upload_form.is_valid():
            import_job = ImportJob(
                school_term=upload_form.cleaned_data["school_term"],
                template=upload_form.cleaned_data["template"],
                data_file=request.FILES["csv"],
            )
            import_job.save()

            result = import_csv.delay(
                import_job.pk,
            )

            return render_progress_page(
                request,
                result,
                title=_("Progress: Import data from CSV"),
                progress_title=_("Import objects …"),
                success_message=_("The import was done successfully."),
                error_message=_("There was a problem while importing data."),
                back_url=reverse("csv_import"),
            )

    context["upload_form"] = upload_form

    return render(request, "csv_import/csv_import.html", context)


class ImportTemplateListView(PermissionRequiredMixin, SingleTableMixin, ListView):
    """Table of all school terms."""

    model = ImportTemplate
    table_class = ImportTemplateTable
    permission_required = "csv_import.view_importtemplate_rule"
    template_name = "csv_import/import_template/list.html"


class ImportTemplateUploadView(PermissionRequiredMixin, FormView):
    form_class = ImportTemplateUploadForm
    permission_required = "csv_import.upload_importtemplate_rule"
    template_name = "csv_import/import_template/upload.html"

    def form_valid(self, form):
        template_defs = form.cleaned_data["template"]
        try:
            ImportTemplate.update_or_create_templates(template_defs)
        except Exception as e:
            messages.error(
                self.request, _("The import of the import templates failed: \n {}").format(e)
            )
            return self.form_invalid(form)
        messages.success(self.request, _("The import of the import templates was successful."))
        return redirect("import_templates")



import pytest

from aleksis.apps.csv_import.util.class_range_helpers import (
    get_classes_per_grade,
    get_classes_per_short_name,
    get_grade_and_class_from_class_range,
    get_max_class,
    get_min_class,
    parse_class_range,
)
from aleksis.core.models import Group

pytestmark = pytest.mark.django_db

CLASSES = [
    "5a",
    "5b",
    "5c",
    "5d",
    "6a",
    "6b",
    "6c",
    "6d",
    "7a",
    "7b",
    "7c",
    "7d",
    "8a",
    "8b",
    "8c",
    "8d",
    "9a",
    "9b",
    "9c",
    "9d",
    "Ea",
    "Eb",
    "Ec",
    "Ed",
    "Q1a",
    "Q1b",
    "Q1c",
    "Q1d",
    "Q2a",
    "Q2b",
    "Q2c",
    "Q2d",
]
CLASSES_PER_GRADE = {
    "5": ["a", "b", "c", "d"],
    "6": ["a", "b", "c", "d"],
    "7": ["a", "b", "c", "d"],
    "8": ["a", "b", "c", "d"],
    "9": ["a", "b", "c", "d"],
    "E": ["a", "b", "c", "d"],
    "Q1": ["a", "b", "c", "d"],
    "Q2": ["a", "b", "c", "d"],
}
CLASSES_PER_GRADE_INCOMPLETE = {
    "5": ["a", "b", "c", "d"],
    "6": ["a", "b", "c", "d"],
    "7": ["a", "b", "c", "d"],
    "8": ["a", "b"],
    "9": ["a", "b", "c", "d"],
    "E": ["a", "b", "c", "d"],
    "Q1": ["a", "b", "c", "d"],
    "Q2": ["b", "c"],
}


def test_get_classes_per_grade():
    assert get_classes_per_grade(CLASSES) == CLASSES_PER_GRADE
    assert get_classes_per_grade(["5a", "5b", "5c", "6a", "6b", "6c", "6d"]) == {
        "5": ["a", "b", "c"],
        "6": ["a", "b", "c", "d"],
    }
    assert get_classes_per_grade([]) == {}


def test_get_min_max_class():
    assert get_min_class(CLASSES_PER_GRADE_INCOMPLETE, "5") == "a"
    assert get_max_class(CLASSES_PER_GRADE_INCOMPLETE, "5") == "d"
    assert get_min_class(CLASSES_PER_GRADE_INCOMPLETE, "8") == "a"
    assert get_max_class(CLASSES_PER_GRADE_INCOMPLETE, "8") == "b"
    assert get_min_class(CLASSES_PER_GRADE_INCOMPLETE, "Q2") == "b"
    assert get_max_class(CLASSES_PER_GRADE_INCOMPLETE, "Q2") == "c"


def test_get_grade_and_class_from_class_range():
    assert get_grade_and_class_from_class_range(CLASSES_PER_GRADE, "5-Q2") == (
        "5",
        "a",
        "Q2",
        "d",
    )
    assert get_grade_and_class_from_class_range(CLASSES_PER_GRADE, "5a-7b") == (
        "5",
        "a",
        "7",
        "b",
    )
    assert get_grade_and_class_from_class_range(CLASSES_PER_GRADE, "5a-d") == (
        "5",
        "a",
        "5",
        "d",
    )
    assert get_grade_and_class_from_class_range(CLASSES_PER_GRADE, "7-8") == (
        "7",
        "a",
        "8",
        "d",
    )


def test_get_classes_per_short_name():
    Group.objects.bulk_create([Group(short_name=name, name=name) for name in CLASSES])

    classes_per_short_name = get_classes_per_short_name(None)

    for class_ in CLASSES:
        assert class_ in classes_per_short_name
        assert classes_per_short_name[class_].short_name == class_


def test_parse_class_range():
    Group.objects.bulk_create([Group(short_name=name, name=name) for name in CLASSES])

    classes_per_short_name = get_classes_per_short_name(None)

    classes = parse_class_range(classes_per_short_name, CLASSES_PER_GRADE, "5-Q2")
    assert sorted([x.short_name for x in classes]) == CLASSES

    classes = parse_class_range(classes_per_short_name, CLASSES_PER_GRADE, "5a-d")
    assert sorted([x.short_name for x in classes]) == ["5a", "5b", "5c", "5d"]



# Generated by Django 3.2.11 on 2022-01-28 19:36

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('csv_import', '0006_from_default_field'),
    ]

    operations = [
        migrations.AddField(
            model_name='importtemplatefield',
            name='virtual_tmpl',
            field=models.TextField(blank=True, null=True, verbose_name='Django template to generate virtual field from'),
        ),
        migrations.AddField(
            model_name='importtemplatefield',
            name='virtual',
            field=models.BooleanField(default=False, verbose_name='Virtual field'),
        ),
    ]



from datetime import date

import pytest
from phonenumbers import PhoneNumber

from aleksis.apps.csv_import.util.converters import (
    parse_comma_separated_data,
    parse_date,
    parse_phone_number,
    parse_sex,
)
from aleksis.core.util.core_helpers import get_site_preferences

pytestmark = pytest.mark.django_db


def test_parse_phone_number():
    get_site_preferences()["internationalisation__phone_number_country"] = "DE"
    fake_number = PhoneNumber(country_code=49, national_number=1635550217)
    assert parse_phone_number("+49-163-555-0217") == fake_number
    assert parse_phone_number("+491635550217") == fake_number
    assert parse_phone_number("0163-555-0217") == fake_number
    assert parse_phone_number("01635550217") == fake_number
    get_site_preferences()["internationalisation__phone_number_country"] = "GB"
    assert parse_phone_number("0163-555-0217") != fake_number
    assert parse_phone_number("01635550217") != fake_number


def test_parse_phone_number_none():
    assert parse_phone_number("") == ""
    assert parse_phone_number("foo") == ""


def test_parse_sex():
    assert parse_sex("w") == "f"
    assert parse_sex("W") == "f"
    assert parse_sex("m") == "m"
    assert parse_sex("M") == "m"
    assert parse_sex("weiblich") == "f"
    assert parse_sex("Weiblich") == "f"
    assert parse_sex("männlich") == "m"
    assert parse_sex("Männlich") == "m"


def test_parse_sex_none():
    assert parse_sex("") == ""
    assert parse_sex("foo") == ""


def test_parse_dd_mm_yyyy():
    get_site_preferences()["csv_import__date_languages"] = "de"
    assert parse_date("12.01.2020") == date(2020, 1, 12)
    assert parse_date("12.12.1912") == date(1912, 12, 12)
    get_site_preferences()["csv_import__date_languages"] = "en-US"
    assert parse_date("12.01.2020") != date(2020, 1, 12)
    assert parse_date("12.12.1912") != date(1912, 12, 12)


def test_parse_dd_mm_yyyy_none():
    assert parse_date("") is None
    assert parse_date("foo") is None
    assert parse_date("12.143.1912") is None


def test_parse_date_iso():
    get_site_preferences()["csv_import__date_languages"] = ""
    assert parse_date("2020-11-12") == date(2020, 11, 12)


def test_parse_date_en():
    get_site_preferences()["csv_import__date_languages"] = ""
    assert parse_date("11/12/2020") == date(2020, 11, 12)
    get_site_preferences()["csv_import__date_languages"] = "de"
    assert parse_date("11/12/2020") != date(2020, 11, 12)


def test_parse_comma_separated_data():
    assert parse_comma_separated_data("1,2,3") == ["1", "2", "3"]
    assert parse_comma_separated_data("1,1") == ["1", "1"]
    assert parse_comma_separated_data(",1") == ["1"]
    assert parse_comma_separated_data("1") == ["1"]
    assert parse_comma_separated_data(" 2, 3") == [" 2", " 3"]


def test_parse_comma_separated_data_none():
    assert parse_comma_separated_data(",") == []
    assert parse_comma_separated_data("") == []



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aleksis',
 'aleksis.apps.csv_import',
 'aleksis.apps.csv_import.management.commands',
 'aleksis.apps.csv_import.migrations',
 'aleksis.apps.csv_import.tests',
 'aleksis.apps.csv_import.tests.models',
 'aleksis.apps.csv_import.tests.util',
 'aleksis.apps.csv_import.util']

package_data = \
{'': ['*'],
 'aleksis.apps.csv_import': ['locale/ar/LC_MESSAGES/*',
                             'locale/de_DE/LC_MESSAGES/*',
                             'locale/fr/LC_MESSAGES/*',
                             'locale/la/LC_MESSAGES/*',
                             'locale/nb_NO/LC_MESSAGES/*',
                             'locale/tr_TR/LC_MESSAGES/*',
                             'templates/csv_import/*',
                             'templates/csv_import/import_template/*']}

install_requires = \
['aleksis-core>=2.7.3,<3.0.0',
 'chardet>=4.0.0,<5.0.0',
 'dateparser>=1.0.0,<2.0.0',
 'pandas>=1.0.0,<2.0.0',
 'phonenumbers>=8.10,<9.0',
 'ruamel.yaml>=0.17.19,<0.18.0',
 'tqdm>=4.62.3,<5.0.0']

entry_points = \
{'aleksis.app': ['csv_import = aleksis.apps.csv_import.apps:CSVImportConfig']}

setup_kwargs = {
    'name': 'aleksis-app-csvimport',
    'version': '2.2.1',
    'description': 'AlekSIS (School Information System)\u200a—\u200aApp for CSV import',
    'long_description': 'AlekSIS (School Information System)\u200a—\u200aApp for CSV imports\n====================================================================\n\nAlekSIS\n-------\n\nThis is an application for use with the `AlekSIS®`_ platform.\n\nFeatures\n--------\n\nThis app provides general CSV imports functions to interact with school administration software.\n\n* Generic and customisable importer based on templates\n* Register import templates in the frontend\n\nSupported systems:\n\n* Schild-NRW (North Rhine-Westphalia, Germany)\n* Pedasos (Schleswig-Holstein, Germany\n\nLicence\n-------\n\n::\n\n  Copyright © 2019, 2020, 2022 Dominik George <dominik.george@teckids.org>\n  Copyright © 2020, 2021 Jonathan Weth <dev@jonathanweth.de>\n  Copyright © 2019 mirabilos <thorsten.glaser@teckids.org>\n  Copyright © 2019 Tom Teichler <tom.teichler@teckids.org>\n\n  Licenced under the EUPL, version 1.2 or later, by Teckids e.V. (Bonn, Germany).\n\nPlease see the LICENCE.rst file accompanying this distribution for the\nfull licence text or on the `European Union Public Licence`_ website\nhttps://joinup.ec.europa.eu/collection/eupl/guidelines-users-and-developers\n(including all other official language versions).\n\nTrademark\n---------\n\nAlekSIS® is a registered trademark of the AlekSIS open source project, represented\nby Teckids e.V. Please refer to the `trademark policy`_ for hints on using the trademark\nAlekSIS®.\n\n.. _AlekSIS®: https://edugit.org/AlekSIS/AlekSIS\n.. _European Union Public Licence: https://eupl.eu/\n.. _trademark policy: https://aleksis.org/pages/about\n',
    'author': 'Dominik George',
    'author_email': 'dominik.george@teckids.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://aleksis.org/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from django.urls import path

from . import views

urlpatterns = [
    path("import", views.csv_import, name="csv_import"),
    path("templates/", views.ImportTemplateListView.as_view(), name="import_templates"),
    path(
        "templates/upload/", views.ImportTemplateUploadView.as_view(), name="upload_import_template"
    ),
]



# Generated by Django 3.2.3 on 2021-05-19 18:41

import django.contrib.sites.managers
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('sites', '0002_alter_domain_unique'),
        ('core', '0017_dashboardwidget_broken'),
        ('csv_import', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ImportJob',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
                ('data_file', models.FileField(upload_to='csv_import/')),
                ('school_term', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='import_jobs', to='core.schoolterm', verbose_name='School term')),
                ('site', models.ForeignKey(default=1, editable=False, on_delete=django.db.models.deletion.CASCADE, to='sites.site')),
                ('template', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='import_jobs', to='csv_import.importtemplate', verbose_name='Import template')),
            ],
            options={
                'verbose_name': 'Import job',
                'verbose_name_plural': 'Import jobs',
            },
            managers=[
                ('objects', django.contrib.sites.managers.CurrentSiteManager()),
            ],
        ),
    ]



import os

from django.contrib.contenttypes.models import ContentType

import pytest

from aleksis.apps.csv_import.default_templates import update_or_create_templates_from_yaml
from aleksis.apps.csv_import.models import ImportTemplate, ImportTemplateField
from aleksis.core.models import Person

pytestmark = pytest.mark.django_db


TEST_FILE_1 = """
test_template:
  model: core.Person
  verbose_name: 'Pedasos: Teachers'
  extra_args:
    has_header_row: true
    separator: "\t"
  fields:
    - short_name
    - last_name
    - first_name
    - date_of_birth
    - sex
    - field_type: departments
      arg_1: foo
      arg_2: 2
    - ignore

"""

TEST_FILE_2 = """
test_template:
  model: core.Person
  verbose_name: 'Pedasos: Teacher'
  extra_args:
    has_header_row: false
    separator: "\t"
  fields:
    - last_name
    - short_name
    - first_name
    - field_type: sex
      arg_1: baz
      arg_2: 3
    - departments
    - ignore
"""


def test_load_sample_template(tmp_path):
    test_filename = os.path.join(tmp_path, "templates.yaml")
    with open(test_filename, "w") as f:
        f.write(TEST_FILE_1)

    assert ImportTemplate.objects.all().count() == 0

    update_or_create_templates_from_yaml(test_filename)

    assert ImportTemplate.objects.all().count() == 1
    assert ImportTemplateField.objects.all().count() == 7
    template = ImportTemplate.objects.all()[0]

    assert template.content_type == ContentType.objects.get_for_model(Person)
    assert template.name == "test_template"
    assert template.verbose_name == "Pedasos: Teachers"
    assert template.has_header_row
    assert template.separator == "\t"

    # Normal field
    field_1 = ImportTemplateField.objects.get(template=template, field_type="short_name")
    assert field_1.index == 0
    assert field_1.args == {}

    # Field with args
    field_2 = ImportTemplateField.objects.get(template=template, field_type="departments")
    assert field_2.index == 5
    assert field_2.args == {"arg_1": "foo", "arg_2": 2}

    # NEW IMPORT

    test2_filename = os.path.join(tmp_path, "templates2.yaml")
    with open(test2_filename, "w") as f:
        f.write(TEST_FILE_2)

    update_or_create_templates_from_yaml(test2_filename)

    assert ImportTemplate.objects.all().count() == 1
    assert ImportTemplateField.objects.all().count() == 6

    template = ImportTemplate.objects.all()[0]
    assert template.name == "test_template"
    assert template.verbose_name == "Pedasos: Teacher"

    # Normal field
    field_1 = ImportTemplateField.objects.get(template=template, field_type="last_name")
    assert field_1.index == 0
    assert field_1.args == {}

    # Normal field removed
    with pytest.raises(ImportTemplateField.DoesNotExist):
        ImportTemplateField.objects.get(template=template, field_type="date_of_birth")

    # Field with args (newly added)
    field_2 = ImportTemplateField.objects.get(template=template, field_type="sex")
    assert field_2.index == 3
    assert field_2.args == {"arg_1": "baz", "arg_2": 3}

    # Field with args (removed)
    field_3 = ImportTemplateField.objects.get(template=template, field_type="departments")
    assert field_3.index == 4
    assert field_3.args == {}



"""Support for parsing typical German class ranges."""

import re
from collections import OrderedDict
from typing import Dict, List, Sequence, Tuple

from aleksis.core.models import Group

REGEX_CLASS_RANGE = r"^([0-9]+|E|Q1|Q2)([a-z]?)-((?:[0-9]+|E|Q1|Q2)?)([a-z]?)"
REGEX_CLASS = r"^([0-9]+|E|Q1|Q2)([a-z])$"
REGEX_CLASS_DB = r"^([0-9]+|E|Q1|Q2)([a-z])$"


def get_classes_per_grade(classes: Sequence[str]):
    """Get classes sorted by grades.

    :param classes: List with all classes
    :return: Dict with grades as keys and class labels as items

    >>> get_classes_per_grade(["5a", "5b", "5c", "6a", "6b", "6c", "6d"])
    {'5': ['a', 'b', 'c'],
    '6': ['a', 'b', 'c', 'd']}
    """
    classes_per_grade = OrderedDict()
    for class_ in classes:
        match = re.match(REGEX_CLASS, class_)

        grade, group = match.group(1), match.group(2)

        if grade not in classes_per_grade:
            classes_per_grade[grade] = []

        if group not in classes_per_grade[grade]:
            classes_per_grade[grade].append(group)

    return classes_per_grade


def get_classes_per_short_name(school_term):
    """Get all groups which match the class range schema and group them by their short names."""
    qs = Group.objects.filter(short_name__regex=REGEX_CLASS_DB, school_term=school_term)

    return {obj.short_name: obj for obj in qs}


def get_min_class(classes_per_grade: Dict[str, Sequence[str]], grade: str):
    """Get label of first class for a grade."""
    return classes_per_grade[grade][0]


def get_max_class(classes_per_grade: Dict[str, Sequence[str]], grade: str):
    """Get label of last class for a grade."""
    return classes_per_grade[grade][-1]


def get_grade_and_class_from_class_range(
    classes_per_grade: Dict[str, Sequence[str]], class_range: str
) -> Tuple[str, str, str, str]:
    """Parse a class range into single parts (grade, class labels) and auto-fill missing parts.

    :param classes_per_grade: All available class labels grouped by grades
    :param class_range: Range to parse
    :return: Start grade, start class label, stop grade, stop class label
    """
    match = re.match(REGEX_CLASS_RANGE, class_range)

    grade_start = match.group(1)
    class_start = match.group(2)
    grade_stop = match.group(3)
    class_stop = match.group(4)

    grade_stop = grade_stop if grade_stop else grade_start

    class_start = class_start if class_start else get_min_class(classes_per_grade, grade_start)
    class_stop = class_stop if class_stop else get_max_class(classes_per_grade, grade_stop)

    return (grade_start, class_start, grade_stop, class_stop)


def parse_class_range(
    classes_per_short_name: Dict[str, Group],
    classes_per_grade: Dict[str, Sequence[str]],
    class_range: str,
) -> List[Group]:
    """Parse a class range into AlekSIS groups.

    :param classes_per_short_name: All available classes in an `OrderedDict` (key is short name)
    :param classes_per_grade: All available class labels grouped by grades
    :param class_range: Range to parse
    :return: List of AlekSIS groups
    """
    if re.match(REGEX_CLASS, class_range):
        # Single class
        return [classes_per_short_name[class_range]]
    else:
        # Class range
        (
            grade_start,
            group_start,
            grade_stop,
            group_stop,
        ) = get_grade_and_class_from_class_range(classes_per_grade, class_range)

        class_a = f"{grade_start}{group_start}"
        class_b = f"{grade_stop}{group_stop}"

        classes = list(classes_per_short_name.keys())
        i_a = classes.index(class_a)
        i_b = classes.index(class_b)

        # Get all classes in this range
        classes_tuple = list(classes_per_short_name.items())[i_a : i_b + 1]

        return [class_t[1] for class_t in classes_tuple]
